from geodeticengine import CoordTrans, CrsSearch, CtSearch

### Example 1
print("Engineering")
points = [[8911.832,5139.165]]
crs_from = "EQUINOR:4100002"
ct_from = "EQUINOR:3100002"
crs_to = "EPSG:25832"
ct = CoordTrans(crs_from=crs_from, crs_to=crs_to, ct_from=ct_from, points=points)
print(f"from:{ct.transform_pointlist()}")
points_to=[[283341.96397220856,6748519.897517173]]
ct = CoordTrans(crs_from=crs_to, crs_to=crs_from, ct_to=ct_from, points=points_to)
print(f"from:{ct.transform_pointlist()}")

### Example 2
points = [[10, 60]]
crs_from = "EPSG:4230"
crs_to = "EPSG:4326"
ct_from = "EPSG:1612"

# Transform coordinates
ct = CoordTrans(crs_from=crs_from, crs_to=crs_to, ct_from=ct_from, points=points)
print(f"Transformed coordinates:{ct.transform_pointlist()}")

# Get transformation pipeline
pipeline = ct.get_pipeline()
print(f"Transformation pipeline: {pipeline}")

### Example 3
points = [[9,65],[12,70]]
crs_from = "ST_ED50_T1133"
crs_to = "ST_WGS84_G4326"

# Transform coordinates
ct = CoordTrans(crs_from=crs_from, crs_to=crs_to, points=points)
print(f"Transformed coordinates:{ct.transform_pointlist()}")

# Get transformation pipeline
pipeline =  ct.get_pipeline()
print(f"Transformation pipeline: {pipeline}")


# Search for projected CRS
crs_query = CrsSearch(types=["bound projected","projected"])
print(crs_query.get_entities())

# Search all transformations between ED50 and WGS 84
ct_query = CtSearch(types=["transformation","concatenated operation"], source_crs="ST_ED50_G4230", target_crs="ST_WGS84_G4326")
print(ct_query.get_entities())


__version__ = "1.0.13"
from .MsalCredential import MsalCredential
from .AuthBearerToken import AuthBearerToken
from .CoordTrans import CoordTrans
from .CrsSearch import CrsSearch
from .CtSearch import CtSearch
from .ege_api import EGE


from dotenv import load_dotenv
load_dotenv(override=True)

import os
from .MsalCredential import MsalCredential
import re

class AuthBearerToken:
    def __init__(self):
        separator = ';'

        client_envs = os.getenv("EGE_CLIENT_ENVS")
        subs_keys_variable = os.getenv("EGE_SUBS_KEYS")
        api_env_variable = os.getenv('EGE_API_ENV', 'prod')

        client_envs_list = ['prod', 'test']
        resource_ids_list = ['c25018ed-b371-4b37-ba4e-4d902aee2b6c','84c8d2ec-6294-47aa-8fd4-f69c870faa3a']
        if not client_envs and ';' not in subs_keys_variable:
            client_envs_list = [api_env_variable]
        if len(client_envs_list)==1:
            if 'prod' in client_envs_list:
                resource_ids_list = ['c25018ed-b371-4b37-ba4e-4d902aee2b6c']
            else:
                resource_ids_list = ['84c8d2ec-6294-47aa-8fd4-f69c870faa3a']

        client_envs = client_envs.split(separator) if client_envs else client_envs_list

        client_ids = os.getenv("EGE_CLIENT_IDS")
        client_ids = client_ids.split(separator) if client_ids else ['c25018ed-b371-4b37-ba4e-4d902aee2b6c','84c8d2ec-6294-47aa-8fd4-f69c870faa3a']

        resource_ids_variable = os.getenv("EGE_RESOURCE_IDS")
        resource_ids = resource_ids_variable.split(separator) if resource_ids_variable else resource_ids_list

        client_secrets_variable =  os.getenv("EGE_CLIENT_SECRETS")
        client_secrets = client_secrets_variable.split(separator) if client_secrets_variable else [None] * len(client_envs)

        subs_keys = subs_keys_variable.split(separator) if subs_keys_variable else [None] * len(client_envs)

        tenant_id = '3aa4a235-b6e2-48d5-9195-7fcf05b459b0'
        authority = f"https://login.microsoftonline.com/{tenant_id}"


        self._credentials = {
            env: MsalCredential(client_id, authority, client_secret)
                .with_scopes([f"{resource_id}/.default openid"]).with_subscription_key(subs_key)
            for env, client_id, client_secret, resource_id,subs_key in zip(client_envs, client_ids, client_secrets, resource_ids, subs_keys)
        }

    def get_authorization(self, environment) -> str:
        cred = self._credentials.get(environment)
        if not cred:
            raise KeyError(f"Failed to retrieve msal client for environment {environment}")
        return cred.get_authorization()

    def authorization_headers(self, uri, headers={}):
        headers |= {"Authorization": self.get_authorization(uri)}
        subs_key = self._credentials.get(uri).get_subscription_key()
        if subs_key:
            headers |= {"Ocp-Apim-Subscription-Key": subs_key}
        return headers





"""
This module contains a python class for getting an access token for the Geodetic Engine
"""
import os
import datetime
from os.path import expanduser
import msal
import sys
import ctypes


class MsalCredential:
    def __init__(self, client_id: str, authority: str, client_secret: str = None, token_cache: str = None):

        self._token_expire = datetime.datetime.now()

        self._client_secret = client_secret
        self._cache = msal.SerializableTokenCache()

        if token_cache:
            self._token_cache = token_cache
        else:
            self._token_cache = expanduser("~/.ege-token-cache-client-"+client_id)
        if os.path.exists(self._token_cache):
            with open(self._token_cache, "r", encoding="utf8") as cachefile:
                self._cache.deserialize(cachefile.read())

        if self._client_secret:
            self._app = msal.ConfidentialClientApplication(client_id, client_credential=client_secret, authority=authority, token_cache=self._cache)
        else:
            self._app = msal.PublicClientApplication(client_id, authority=authority, token_cache=self._cache)

        self._scopes = None
        self._subscription_key = None


    def with_scopes(self, scopes):
        self._scopes = scopes
        return self

    def with_subscription_key(self, subs_key):
        self._subscription_key = subs_key
        return self

    def _has_gui(self):
        if sys.platform == "win32":
            try:
                ctypes.windll.user32.GetSystemMetrics(0)
                return True
            except:
                return False
        else:
            try:
                os.environ["DISPLAY"]
                return True
            except:
                return False


    def renew_token(self) -> dict:
        token_now = datetime.datetime.now()

        result = None
        if self._client_secret:
            result = self._app.acquire_token_for_client(self._scopes)
        else:
            accounts = self._app.get_accounts()
            if accounts:
                result = self._app.acquire_token_silent(
                    self._scopes, account=accounts[0])
            if not result:
                if self._has_gui():
                    result = self._app.acquire_token_interactive(self._scopes)
                else:
                    flow = self._app.initiate_device_flow(self._scopes)
                    if "user_code" not in flow:
                        raise Exception("Failed to create device flow")
                    print(flow["message"])
                    result = self._app.acquire_token_by_device_flow(flow)
            self._save_cache_if_changed()

        if "access_token" in result:
            self._token_expire = token_now + datetime.timedelta(seconds=result["expires_in"])
            self._token = {
                'token_type' : result["token_type"],
                'token': result['access_token']
            }
        else:
            print(result)
            raise Exception("No access token in result")

    def get_authorization(self) -> str:
        diff = self._token_expire - datetime.datetime.now()
        if diff.total_seconds() < 5.0:
            self.renew_token()
        return f'{self._token["token_type"]} {self._token["token"]}'

    def get_subscription_key(self) -> str:
        return self._subscription_key

    def _save_cache_if_changed(self):
        if self._cache.has_state_changed:
            with open(self._token_cache, "w", encoding="utf8") as cache_file:
                cache_file.write(self._cache.serialize())


"""
Module for running automated tests on the pytyhon package using a test data set with expected values transformed using Blue Marbel's Geographic Calculator.
The test data set is downloaded from Azure Blob Store
"""

import os
import sys
import pytest
import ast
import pandas as pd
import numpy as np
from pygeodesy.ellipsoidalVincenty import LatLon
from pyproj import CRS
from azure.storage.blob import BlobServiceClient
from azure.identity import DefaultAzureCredential
import io
from os import environ as env
from dotenv import load_dotenv
load_dotenv(override=True)

path_to_api = os.path.abspath(os.path.join(os.path.dirname(__file__), '../../../api'))
path_to_pypack = os.path.abspath(os.path.join(os.path.dirname(__file__), '../'))
sys.path.append(path_to_api)
sys.path.append(path_to_pypack)

from geodeticengine import CoordTrans

CONTAINER_NAME = 'test-data'
blob_name = 'test_data_ege.csv'

# Download the test data set from Azure Blob store
CONTAINER_NAME = 'test-data'
blob_name = 'test_data_ege.csv'

connection_string = env.get("AZ_CONNECTION_STRING")
if connection_string and connection_string.startswith("https"):
    blob_service = BlobServiceClient(account_url=connection_string, credential=DefaultAzureCredential())
else:
    blob_service = BlobServiceClient.from_connection_string(connection_string)

blob = blob_service.get_blob_client(CONTAINER_NAME, blob_name)
downloaded_data = blob.download_blob().readall().decode('utf-8')
# Create a DataFrame
DF_ALL = pd.read_csv(io.StringIO(downloaded_data), delimiter=';')


### - Run through the test dataset and compaire agains Geographic Calculator
def get_dist_on_ellipsoid_between_points(transformed_coords: np.ndarray, reference_points: np.ndarray) -> float:
    """
    Find the distance on the ellipsoid between two points defined
    by geographic coordinates.
    """
    lon = transformed_coords[:,0]; lat = transformed_coords[:,1]
    lon_ref = reference_points[:,0]; lat_ref = reference_points[:,1]
    diff_lst = []
    for p in range(len(transformed_coords)):
        if (lat[p] == 0 and lon[p]== 0 and lat_ref[p] == 0 and lon_ref[p] == 0) \
            or ((abs(lat[p])==90 or abs(lon[p])==180) and (abs(lat_ref[p])==90 or abs(lon_ref[p])==180)): # check if they are all zero (to prevent pygeodesy.ellipsoidalVincenty.VincentyError: ambiguous: LatLon(00°00′00.0″N, 000°00′00.0″E) antipodal to  LatLon(00°00′00.0″N, 000°00′00.0″E)
            diff_lst.append(0)
            continue
        point_transformed = LatLon(lat[p], lon[p])
        point_reference = LatLon(lat_ref[p], lon_ref[p])
        diff_lst.append(point_transformed.distanceTo(point_reference))
    return np.array(diff_lst)


# # Build up the payload
df = DF_ALL[DF_ALL["CRS_Type"] != "engineering"]
PAYLOAD_XY = []
for idx, row in df.iterrows():
    # Extract the data from the current row
    coords_to_transform = ast.literal_eval(row['Expected_Results'])
    crs_from = "EPSG:4326"
    crs_to = row['CRS_Code']
    ct_authcode = row['Transformation_Code']
    crs_type = row['CRS_Type']
    expected_coords = ast.literal_eval(row['Points_To_Transform'])
    error_threshold = row['Error_Threshold']
    # Build up the payload tuple
    payload = ({"coordinates_from": [coords_to_transform], "crs_from": crs_from, "crs_to": crs_to,
                 "ct_authcode": ct_authcode, "crs_type":crs_type, "error_threshold": error_threshold }, expected_coords)
    PAYLOAD_XY.append(payload)

@pytest.mark.filterwarnings("ignore::DeprecationWarning")
@pytest.mark.parametrize("payload, expected_coords ", PAYLOAD_XY)
def test_coordinate_transform(payload, expected_coords):
    param_dict = payload
    crs_from = param_dict["crs_from"]
    crs_to = param_dict["crs_to"]
    ct_authcode = param_dict["ct_authcode"]
    crs_type = param_dict["crs_type"]
    error_threshold = param_dict["error_threshold"]
    if not isinstance(ct_authcode,str) and np.isnan(ct_authcode):
        ct_authcode = None
    lst_of_point = param_dict["coordinates_from"][0]
    # Transform first step (Projceted CRS -> WGS84)
    ct_geog = CoordTrans(crs_from=crs_to, crs_to=crs_from, ct_from=ct_authcode, points=lst_of_point)
    transformed_coords_geog = ct_geog.transform_pointlist()
    if CRS.from_user_input(crs_from).is_geographic:
        diff = get_dist_on_ellipsoid_between_points(np.array(transformed_coords_geog), np.array(expected_coords))
    else:
        diff = np.abs(np.array(transformed_coords_geog) - np.array(expected_coords))
    assert (diff < error_threshold).all() == True, f"Test case failed for CRS: {crs_to}, transformation:{ct_authcode}, max difference: {np.round(np.max(diff),3)}, threshold: {error_threshold}"
    # Transform second step (WGS84 -> Projceted CRS)
    ct = CoordTrans(crs_from=crs_from, crs_to=crs_to, ct_to=ct_authcode, points=transformed_coords_geog)
    transformed_coords = np.array(ct.transform_pointlist())
    # Compare agaist the input coordinates
    if 'geographic' in crs_type:
        diff = get_dist_on_ellipsoid_between_points(transformed_coords, np.array(lst_of_point))
    else:
        diff = np.abs(transformed_coords - np.array(lst_of_point))
    # Check if difference is greater than threshold value
    assert (diff < error_threshold).all() == True, f"Test case failed for CRS: {crs_to}, transformation:{ct_authcode}, max difference: {np.round(np.max(diff),3)}, threshold: {error_threshold}"




## Test engineering
df_engineering = DF_ALL[DF_ALL["CRS_Type"]=="engineering"]
PAYLOAD_ENGINEERING = []
for idx, row in df_engineering.iterrows():
    # Extract the data from the current row
    coords_to_transform = ast.literal_eval(row['Points_To_Transform'])
    crs_from = row['CRS_Code']
    crs_to = row['ENG_CRS_To']
    ct_authcode = row['Transformation_Code']
    expected_coords = ast.literal_eval(row['Expected_Results'])
    error_threshold = row['Error_Threshold']
    # Build up the payload tuple
    payload = ({"points_to_transform": [coords_to_transform], "crs_from": crs_from, "crs_to": crs_to,
                 "ct_authcode": ct_authcode, "error_threshold": error_threshold }, expected_coords)
    PAYLOAD_ENGINEERING.append(payload)

@pytest.mark.filterwarnings("ignore::DeprecationWarning")
@pytest.mark.parametrize("payload_engineering, expected_coords ", PAYLOAD_ENGINEERING)
def test_engineering_transform(payload_engineering, expected_coords):
    param_dict = payload_engineering
    crs_from = param_dict["crs_from"]
    crs_to = param_dict["crs_to"]
    ct_authcode = param_dict["ct_authcode"]
    error_threshold = param_dict["error_threshold"]
    points_to_transform = param_dict["points_to_transform"][0]
    # Transform first step (Engineering CRS -> Global CRS like WGS84)
    ct_eng2glob = CoordTrans(crs_from=crs_from, crs_to=crs_to, ct_from=ct_authcode, points=points_to_transform)
    transformed_coords_glob = ct_eng2glob.transform_pointlist()
    if CRS.from_user_input(crs_to).is_geographic:
        diff = get_dist_on_ellipsoid_between_points(np.array(transformed_coords_glob), np.array(expected_coords))
    else:
        diff = np.abs(np.array(transformed_coords_glob) - np.array(expected_coords))
    assert (diff < error_threshold).all() == True, f"Test case failed for CRS: {crs_to}, transformation:{ct_authcode}, max difference: {np.round(np.max(diff),3)}, threshold: {error_threshold}"
    # Transform second step (Global CRS like WGS84 -> Engineering CRS)
    ct_glob2eng = CoordTrans(crs_from=crs_to, crs_to=crs_from, ct_to=ct_authcode, points=transformed_coords_glob)
    transformed_coords = np.array(ct_glob2eng.transform_pointlist())
    # Compare agaist the input coordinates
    diff = np.abs(transformed_coords - np.array(points_to_transform))
    # Check if difference is greater than threshold value
    assert (diff < error_threshold).all() == True, f"Test case failed for CRS: {crs_to}, transformation:{ct_authcode}, max difference: {np.round(np.max(diff),3)}, threshold: {error_threshold}"




"""
Misc useful functions
"""

import re
from geodeticengine import request_wkt_for_crs

def find_indices_of_attribute(ogc_wkt_str, attribute_name):
    """
    Function that loops through the input string and finds the
    start index and end index of a defined attribute in the WKT-string. Ex: "GEOGCS"
    The function keeps track of the nested brackets to ensure that it finds
    the correct "]]" that closes the defined attribute. This function finds all indices for every occurency of attribute_name

    Parameters
    ----------
    trans_ogc_wkt_str : The transformation WKT-string in the OGC dialect
    attribute_name : The attribute to find the indexes for (ex:"GEOGCS")

    Returns
    -------
    indices = A list of tuples containing start and stop index for every occurency of attribute name
    """
    indices = []
    start_index = ogc_wkt_str.find(f"{attribute_name}[")
    while start_index != -1:
        stack = [] # list for keeping track of the nested brackets
        for i, char in enumerate(ogc_wkt_str[start_index:]):
            if char == "[":
                stack.append(char)
            elif char == "]":
                stack.pop()
                if not stack:
                    end_index = start_index + i + 1
                    indices.append((start_index, end_index))
                    break
        start_index = ogc_wkt_str.find(f"{attribute_name}[", end_index)
    return indices



def extract_target_crscode_from_transWKT(trans_WKT_ogc) -> list:
    """
    Function that extracts EPSG code from the two CRSs that are
    included in the transformation string. Returns a tuple. (ex:('4202', '4326'))
    """
    if not "ENGCRS" in trans_WKT_ogc:
        indices = find_indices_of_attribute(trans_WKT_ogc, 'GEOGCS')
        epsg_codes = []
        for i in range(len(indices)):
            geo_str = trans_WKT_ogc[indices[i][0]:indices[i][1]]
            idx = geo_str.rindex('AUTHORITY[')
            epsg_codes.append(re.findall('[0-9]+', geo_str[idx:-1])[0])
        source_crs, target_crs = epsg_codes
        # return  source_crs, target_crs
        return [source_crs], [target_crs]
    else:
        return extract_target_crs_from_engineering_transWKT(trans_WKT_ogc)



def extract_geogcrs_codes_from_crs_wkt(ogc_wkt):
    """Function that extracts CRS code for the GEOGCS in WKT_GDAL."""
    epsg_codes = []
    for attribute in ['GEOGCS', 'PROJCS', 'LOCAL_CS']:
        indices = find_indices_of_attribute(ogc_wkt, attribute)
        for start, end in indices:
            geo_str = ogc_wkt[start:end]
            idx = geo_str.rindex('AUTHORITY[')
            epsg_code = re.findall('[0-9]+', geo_str[idx:-1])[0]
            epsg_codes.append(epsg_code)
    return epsg_codes


def extract_target_crs_from_engineering_transWKT(trans_WKT_ogc) -> list:
    """
    For engineering CRS only.
    Function that extracts authority code from the two CRSs that are
    included in the transformation string. Returns a tuple. (ex: ('4202', '4326'))
    """
    epsg_codes = []
    for attribute in ['ENGCRS', 'PROJCS']:
        indices = find_indices_of_attribute(trans_WKT_ogc, attribute)
        for start, end in indices:
            geo_str = trans_WKT_ogc[start:end]
            idx = geo_str.rindex('AUTHORITY[')
            epsg_code = re.findall('[0-9]+', geo_str[idx:-1])[0]
            epsg_codes.append(epsg_code)
    source_crs, target_crs = epsg_codes
    geogcs_ind = find_indices_of_attribute(trans_WKT_ogc, 'GEOGCS')
    geo_str = trans_WKT_ogc[geogcs_ind[0][0]:geogcs_ind[0][1]]
    idx = geo_str.rindex('AUTHORITY[')
    geogcs_code = re.findall('[0-9]+', geo_str[idx:-1])[0]
    return [source_crs], [target_crs, geogcs_code]


def get_wkt_gdal_for_crs(crs_authcode) -> str:
    return request_wkt_for_crs(crs_authcode, "WKT1_GDAL")['crs_wkt']


def get_wkt_gdal_for_ct(ct_authcode) -> str:
    # return request_wkt_for_crs(ct_authcode, "WKT1_GDAL")['transformation_wkt']
    return request_wkt_for_crs(ct_authcode, "WKT2_2019")['transformation_wkt']

def determine_trans_dir(crs_authcode, ct_authcode):
    is_inverse = False
    crs_wkt = get_wkt_gdal_for_crs(crs_authcode)
    ct_wkt = get_wkt_gdal_for_ct(ct_authcode)
    crs_base = extract_geogcrs_codes_from_crs_wkt(crs_wkt)
    source_crs, target_crs = extract_target_crscode_from_transWKT(ct_wkt)
    if set(crs_base).intersection(source_crs) and not set(crs_base).intersection(target_crs):
        is_inverse = False
    elif set(crs_base).intersection(target_crs) and not set(crs_base).intersection(source_crs):
        is_inverse = True
    else:
        raise ValueError("The base CRS of the CRS does not match neither the source or target CRS of the transformation.")
    return is_inverse



if __name__ == "__main__":
    is_inverse = determine_trans_dir("ESPG:4230", "EPSG:1612")
    print(is_inverse)


import json
import requests
import os
from functools import wraps
from typing import Literal
from .AuthBearerToken import AuthBearerToken
from dotenv import load_dotenv
load_dotenv(override=True)


class InputException(ValueError):
    pass

class AuthenticationError(ValueError):
    pass

class EndpointException(ValueError):
    pass




def trim_all_whitespace(string):
    return string.strip()

def strip_arguments(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        new_args = tuple(trim_all_whitespace(arg) if isinstance(arg, str) else arg for arg in args)
        new_kwargs = {k: trim_all_whitespace(v) if isinstance(v, str) else v for k, v in kwargs.items()}
        return func(*new_args, **new_kwargs)
    return wrapper




class EGE:
    """
    Singleton Class for communicating with the Geodetic Engine API
    """

    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(EGE, cls).__new__(cls)
            cls.instance.initialize()
        return cls.instance

    def initialize(self):
        self.EGE_API_ENV = os.getenv('EGE_API_ENV', 'prod')  # set "prod" as default if not defined
        if self.EGE_API_ENV in ('test', 'dev'):
            self.GE_API = f"https://api-{self.EGE_API_ENV}.gateway.equinor.com/geodetic-engine/v1"
        else:
            self.GE_API = "https://api.gateway.equinor.com/geodetic-engine/v1"

        self.GE_ENDPOINT_TRANSFORMATION = f"{self.GE_API}/transform"
        self.GE_WKT_ENDPOINT = f'{self.GE_API}/crs/wkt'
        self.GE_VALIDATE_ENDPOINT = f'{self.GE_API}/validate'
        self.GE_CRS_SEARCH_ENDPOINT = f'{self.GE_API}/crs/search'
        self.GE_CT_SEARCH_ENDPOINT = f'{self.GE_API}/ct/search'
        self.auth = AuthBearerToken()

    @strip_arguments
    def get_pipeline(self, crs_from, crs_to):
        """Request transformation pipeline"""
        input_params = {'crs_from': crs_from, 'crs_to': crs_to}
        response = self.call_endpoint(self.GE_ENDPOINT_TRANSFORMATION, 'post', input_params)
        return response

    @strip_arguments
    def transform_crs(self, points, crs_from, crs_to):
        """Transform points"""
        input_params = {'coordinates_from': points, 'crs_from': crs_from, 'crs_to': crs_to}
        response = self.call_endpoint(self.GE_ENDPOINT_TRANSFORMATION, 'post', input_params)
        return response

    @strip_arguments
    def crs_search(self, types, polygon_coords, target_crs):
        """Search CRSs"""
        input_params = {'types': types, 'polygon_coords': polygon_coords, 'target_crs': target_crs}
        response = self.call_endpoint(self.GE_CRS_SEARCH_ENDPOINT, 'post', input_params)
        return response

    @strip_arguments
    def ct_search(self, types, polygon_coords, source_crs, target_crs):
        """Query CTs"""
        input_params = {'types': types, 'polygon_coords': polygon_coords, 'source_crs': source_crs, 'target_crs': target_crs}
        response = self.call_endpoint(self.GE_CT_SEARCH_ENDPOINT, 'post', input_params)
        return response

    def validate_input(self, user_input):
        """
        Validate user input using
        the API validate endpoint
        """
        endpoint = f"{self.GE_VALIDATE_ENDPOINT}?input={user_input}"
        response = self.call_endpoint(endpoint, 'get')
        return response

    @strip_arguments
    def request_wkt_for_crs(self, crs:str, wkt_version: Literal["WKT1_ESRI", "WKT1_GDAL", "WKT2_2019"]):
        """Request WKT for a CRS"""
        url = f"{self.GE_WKT_ENDPOINT}?input={crs}&wkt_version={wkt_version}"
        return self.call_endpoint(url, 'get')

    def call_endpoint(self, endpoint,  http_method='get', input_params_dict=None):
        """Get endpoint together with valid bearer token"""
        try:
            if http_method.lower() == 'get':
                if input_params_dict:
                    response = requests.get(
                        endpoint, headers=self.auth.authorization_headers(self.EGE_API_ENV), params=input_params_dict, timeout=30
                    )
                else:
                    response = requests.get(
                        endpoint, headers=self.auth.authorization_headers(self.EGE_API_ENV), timeout=30
                    )
            elif http_method.lower() == 'post':
                if not input_params_dict:
                    raise InputException("Input parameters are empty")
                json_payload = json.dumps(input_params_dict).strip(' \t\n\r')
                response = requests.post(endpoint, headers=self.auth.authorization_headers(self.EGE_API_ENV), data=json_payload, timeout=30)
            else:
                raise InputException("Invalid HTTP method specified")

            result = response.json()
            response.raise_for_status()
        except requests.exceptions.Timeout:
            raise EndpointException("Request timed out")
        except requests.exceptions.RequestException as e:
            if response.status_code == 401:
                message = result['message']
            else:
                message = result.get('detail', {}).get('reason', '')
            raise EndpointException(f"{e}. {message}")
        except ValueError as e:
            raise EndpointException(f"JSON decoding failed: {e}")
        return result



import numpy as np
from pyproj import Transformer, network
from geodeticengine.ege_api import EGE

network.set_network_enabled(active=True)


class CoordTrans():
    """
    Class for transforming coordinates using Equinor Geodetic Engine.
    Takes in CRS and CT info both using autority code and Equinor alias.

    """
    def __init__(self, crs_from:str, crs_to:str, ct_from:str=None, ct_to:str=None, points:list[float]=None):
        """
        Examples:
        >>> ct = CoordTrans(crs_from="ST_ED50_G4230", crs_to="EPSG:4326", ct_from="EPSG:1612", points=[[10, 60]])
        >>> ct = CoordTrans(crs_from="ST_ED50_UTM31N_P23031_T1613", crs_to="ST_ETRS89_UTM31N_P25831_T1149", points=[[600000, 6714000]])
        >>> ct = CoordTrans(crs_from="EPSG:23031", crs_to="EPSG:25831", ct_from="EPSG:1613", ct_to="EPSG:1149",points=[[600000, 6714000]])
        """
        self.points = points
        self.crs_from = self.__strip_input(crs_from)
        self.crs_to = self.__strip_input(crs_to)
        self.ct_from = ct_from
        self.ct_to = ct_to
        if self.ct_from is not None:
            self.ct_from = self.__strip_input(ct_from)
        if self.ct_to is not None:
            self.ct_to = self.__strip_input(ct_to)

    def validate_input(self, entity):
        """
        Validates the input parameters.

        Examples:
        >>> ct = CoordTrans(crs_from="ST_ED50_G4230", crs_to="EPSG:4326",  ct_from="EPSG:1612", points=[[10,60]])
        >>> ct.validate_input(ct.crs_from)
        [{'code': 4230, 'name': 'ST_ED50_G4230', 'naming_system': 'Equinor', 'authority': 'EPSG', 'uom': 'degree', 'coord_sys': 'ellipsoidal', 'type': 'geographic 2D', 'is_alias': True, 'alias': [{'name': 'ED50', 'naming_system': 'EPSG'}, {'name': 'GCS_European_1950', 'naming_system': 'ESRI'}], 'deprecations': []}]
        """
        response = EGE().validate_input(entity).get('results')
        if response is None:
            raise ValueError(f"{entity} is not a valid entity")
        return response


    def __strip_input(self, user_input):
        """Strips input strings"""
        return user_input.strip().replace(" ","")


    def transform_pointlist(self) -> list[float]:
        """
        Transform a list of points and return a list
        of transformed coordinates

        Examples:
        >>> ct = CoordTrans(crs_from="EPSG:4230", crs_to="EPSG:4326",  ct_from="EPSG:1612", points=[[10,60]])
        >>> ct.transform_pointlist()
        [[9.99860677505385, 59.999554562447024]]
        """

        pipeline = self.get_pipeline()
        trans_coords = self.transform_from_pipeline(self.points, pipeline)
        return trans_coords


    def get_pipeline(self) -> str:
        """
        Return a PROJ transformation pipeline from Equinor Geodetic Engine.
        Examples:
        >>> ct = CoordTrans(crs_from="ST_ED50_G4230", crs_to="EPSG:4326", ct_from="EPSG:1612")
        >>> ct.get_pipeline()
        '+proj=pipeline +step +proj=push +v_3 +step +proj=cart +ellps=intl +step +proj=helmert +x=-116.641 +y=-56.931 +z=-110.559 +rx=0.893 +ry=0.921 +rz=-0.917 +s=-3.52 +convention=position_vector +step +inv +proj=cart +ellps=WGS84 +step +proj=pop +v_3'
        """
        crs_from = self.crs_from
        if self.ct_from:
            crs_from = f"({self.crs_from},{self.ct_from})"

        crs_to = self.crs_to
        if self.ct_to:
            crs_to = f"({self.crs_to},{self.ct_to})"
        response = EGE().get_pipeline(crs_from, crs_to)
        return response['pipeline']


    def transform_from_pipeline(self, coordinates, pipeline) -> list:
        """
        Transform a list of points using PROJ pipeline and return a list
        of transformed coordinates.

        Examples:
        >>> ct = CoordTrans(crs_from="ST_ED50_G4230", crs_to="EPSG:4326",  ct_from="EPSG:1612")
        >>> ct.transform_from_pipeline([[10,60]],ct.get_pipeline())
        [[9.99860677505385, 59.999554562447024]]
        """
        transformer = Transformer.from_pipeline(pipeline)
        np_coordinates = np.array(coordinates, dtype=np.double)
        x, y = np_coordinates.T # extract x and y
        tx, ty = transformer.transform(x, y)
        transformed_coordinates = np.vstack([tx, ty]).T.tolist()

        return transformed_coordinates


if __name__ == "__main__":
    import doctest
    doctest.testmod()


from geodeticengine.ege_api import EGE


class CrsSearch():
    """
    Class to query Equinor named CRSs based on polygon filter, crs types, target crs and usage scopes.
    The polygon filter must be defined with a list of WGS84 longitude, latitude coordinates.

    """
    def __init__(self,types:list[str], target_crs:str=None,
                 polygon_coords:list[float]=None):
        """
        Examples:
        >>> crs_search = CrsSearch(types=["bound projected","projected"], polygon_coords=[[1.278828871805691,58.07568845044884],[3.690287338364835,59.20344381800123],[2.274239008972083,60.12176489296384],[-0.1274790229401068, 59.8722761692493]], target_crs="ST_WGS84_G4326")
        """
        self.types = types
        self.polygon_coords = polygon_coords
        self.target_crs = target_crs

    def get_entities(self):
        return EGE().crs_search(self.types, self.polygon_coords, self.target_crs)

if __name__ == "__main__":
    import doctest
    doctest.testmod()